import datetime
from functools import partial
from math import ceil

from django.contrib.auth.models import User
from django.db import models
from djangotools.common.types import parse_date, date_to_string

from baby.app.models.const import get_average_month, get_min_period, get_max_period


class Cycles(list):

    def __init__(self, start=None, end=None):
        super().__init__()
        self.start = start
        self.end = end

    def iter_dates(self, start=None, end=None):
        if not self: return
        start = start or self.start or self[0].start
        end = end or self.end or self[-1].end
        for cycle in self:
            if cycle.end < start: continue
            for date, cur in cycle:
                if date<start: continue
                if date>end: return
                yield date, cur



class Cycle:
    DEBUT="debut"
    OVULATION="ovulation"
    FERTILITE="fertilite"


    def __init__(self):
        self.start : datetime.date = None
        self.end : datetime.date = None
        self.length : int = None
        self.min_cycle : int = None
        self.max_cycle : int = None
        self.fertility_start_offset : int = None
        self.fertility_start : datetime.date = None
        self.fertility_end_offset : int = None
        self.fertility_end : datetime.date = None
        self.ovulation_offset : int = None
        self.ovulation : datetime.date = None

    def as_list(self, with_date=False):
        if with_date:
            ret = [(self.start + datetime.timedelta(days=i), []) for i in range(self.length)]
            ret[0][1].append(self.DEBUT)
            if self.length >= 14:
                ret[self.length - 14][1].append(self.OVULATION)
            if self.fertility_start_offset:
                [ret[i - 1][1].append(self.FERTILITE) for i in range(self.fertility_start_offset, self.fertility_end_offset + 1)  if 0 < i < self.length]
        else:
            ret = [[] for _ in range(self.length)]
            ret[0].append(self.DEBUT)
            if self.length >= 14:
                ret[self.length-14].append(self.OVULATION)
            if self.fertility_start_offset:
                [ret[i-1].append(self.FERTILITE) for i in range(self.fertility_start_offset, self.fertility_end_offset+1) if 0 > i > self.length]
        return ret

    def __iter__(self):
        return iter(self.as_list(with_date=True))

    def __repr__(self):
        return f"<Cycle {date_to_string(self.start)} to {date_to_string(self.end)} ({self.length})>"

    def __str__(self):
        return self.__repr__()

    @property
    def json(self):
        return {
            "start" : date_to_string(self.start),
            "end" : date_to_string(self.end),
            "length" : self.length,
            "ovulation" : date_to_string(self.ovulation),
            "fertility" : [date_to_string(self.fertility_start),date_to_string(self.fertility_end)],
            "ovulation_offset" : self.ovulation_offset,
            "fertility_offset" : [self.fertility_start_offset, self.fertility_end_offset],
        }


    def _init(self):
        assert self.start
        if not self.end:
            assert self.length
            self.end = self.start + datetime.timedelta(days=self.length)
        elif not self.length:
            assert  self.end
            self.length = (self.end - self.start).days
        else:
            return

        self._fertility_init()

        if self.fertility_start_offset is not None:
            self.fertility_start = self.start + datetime.timedelta(days=self.fertility_start_offset-1)

        if self.fertility_end_offset is not None:
            self.fertility_end = self.start + datetime.timedelta(days=self.fertility_end_offset-1)

        if self.ovulation_offset is not None:
            self.ovulation = self.start + datetime.timedelta(days=self.ovulation_offset-1)

    def _fertility_init(self):
        self.fertility_start_offset = self.min_cycle and int(self.min_cycle - 18)
        self.fertility_end_offset = self.max_cycle and ceil(self.max_cycle - 11)
        self.ovulation_offset = ceil(self.length - 11)

    @classmethod
    def create(cls, user, start, end=None, length=None, monthes=None):
        filter = partial(Regle.objects.filter, user=user)
        kwargs = {
            "date__lte": start
        }

        if monthes != 0:
            monthes = monthes if monthes is not None else get_average_month(user)
            kwargs["date__gte"] =  start - datetime.timedelta(days=monthes*30)

        regles = list(filter(**kwargs ).order_by("date"))
        periods = Regle.objects.rgeles_to_period(regles, min_period=get_min_period(user), max_period=get_max_period(user))
        mini = None
        maxi = None
        if periods:
            mini = min(periods)
            maxi = max(periods)

        cycle = cls()
        cycle.start = start
        if end:
            cycle.end = end
        elif length:
            cycle.length = length
        else:
            raise ValueError

        cycle.min_cycle = mini
        cycle.max_cycle = maxi
        cycle._init()
        return cycle

    @classmethod
    def from_day(cls, user, ref_date=None, monthes=None):
        today = datetime.date.today()
        date = parse_date(ref_date) or today
        regle = Regle.objects.last_regle_before(user, date)
        end = Regle.objects.next_regle_after(user, date)
        monthes = monthes if monthes is not None else get_average_month(user)

        kwargs = {"user": user}
        x = user.pref.regle_max_period
        if ref_date and ref_date > datetime.date.today():
            length = Regle.objects.average_period(user, monthes=monthes)
            ilength = int(length)

            offset =  int((date - regle.date).days / ilength) * int(ilength)
            kwargs["start"] = regle.date + datetime.timedelta(days=int(offset))
            kwargs["end"] = kwargs["start"] + datetime.timedelta(days=ilength)
        elif regle:
            kwargs["start"] = regle.date
            if end:
                kwargs["end"]  = end.date
            else:
                length = Regle.objects.average_period(user, monthes=monthes)
                if length is not None:
                    kwargs["length"]  = int(length)
                else:
                    return None
        else:
            return None

        return cls.create(**kwargs)

    @classmethod
    def from_date_range(cls, user, start, end):
        if start > end: start, end = end, start
        cycles = Cycles(start=start, end=end)
        curr = start
        while curr <= end:
            cycle = cls.from_day(user, curr)
            if  cycle is None:
                break
            cycles.append(cycle)
            curr = cycle.end + datetime.timedelta(days=1)

        return cycles

    @classmethod
    def from_month(cls, user, mont_or_date, year):
        if isinstance(mont_or_date, str):
            mont_or_date = parse_date(mont_or_date)
        if isinstance(mont_or_date, (datetime.date, datetime.datetime)):
            mont_or_date, year = mont_or_date.month, mont_or_date.year
        assert isinstance(mont_or_date, int) and isinstance(year, int)
        if year < 100: year += 2000
        assert 0 < mont_or_date <= 12
        start = datetime.date(year, mont_or_date, 1)
        end = start + datetime.timedelta(days=32)
        end = end - datetime.timedelta(days=end.day)
        return cls.from_date_range(user, start, end)


class RegleManager(models.Manager):
    def last_year(self, user):
        return self.filter(user=user, date__gte=datetime.date.today() - datetime.timedelta(days=-395))

    def create(self, user, date=None):
        date = parse_date(date)
        if date is None:
            date = datetime.date.today()
        return super().create(user=user, date=date)
    def last_regle_before(self, user, date=None):
        if date is None: date = datetime.date.today()
        regle = self.filter(user=user, date__lte=date).order_by("-date")[:1]
        regle = list(regle)
        return regle[0] if regle else None

    def average_period(self, user, monthes=None, date_min=None, date_max=None):
        kwargs = {"user": user}

        if date_min or date_max:
            if date_min: kwargs["date__gte"] = date_min
            if date_max: kwargs["date__lte"] = date_max
        elif monthes:
            kwargs["date__gte"] = datetime.date.today() - datetime.timedelta(days=monthes*30)

        regles = self.filter(**kwargs).order_by("date")
        data = self.rgeles_to_period(regles)
        if data:
            return sum(data) / len(data)
        return None
    @classmethod
    def rgeles_to_period(cls, data, start_date=False, min_period=None, max_period=None):
        data = iter(data)
        periods = []
        try:
            last = next(data)
            while True:
                curr = next(data)
                dt = curr.date - last.date
                if start_date:
                    dt = (last.date, dt)
                if (min_period is not None and dt.days<min_period or
                    max_period is not None and dt.days>max_period):
                    last=curr
                    continue
                periods.append(dt.days)
                last = curr
        except StopIteration:
            pass
        return periods

    def next_regle_after(self, user, date):
        regle = self.filter(user=user, date__gt=date).order_by("date")[:1]
        regle = list(regle)
        return regle[0] if regle else None


class Regle(models.Model):

    objects = RegleManager()
    user = models.ForeignKey(User, on_delete=models.CASCADE)
    date = models.DateField()
    DoesNotExist : Exception

    class Meta:
        unique_together = ["user", "date"]


    def __repr__(self):
        return f"<Regle {self.user.username}  {self.date}>"

    def __str__(self):
        return self.__repr__()